/***************************************************************************
 *   Copyright (C) 2007 by Julien Keller   *
 *   j.keller@imperial.ac.uk   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#define Complex  std::complex<double>
#define _Complex_I std::complex<double>(0, 1)
#define _1  (std::complex<double>(1, 0))
#define _2  (std::complex<double>(2, 0))
#define _0  (std::complex<double>(0, 0))
#define PI  3.14159265358979323846

#define pow2(x) (x*x)
#define pow3(x) (x*x*x)
#define pow4(x) pow2(x*x)
#define pow5(x) (x*pow4(x))
#define pow6(x) pow2(pow3(x))
#define pow8(x) pow2(pow4(x))

#include "plot_geodesic_p2blow3pts.h"

int main(int argc, char **argv)
{
  int a,b,i,j;
  double to; 
  ifstream param("coefficients.dat");
  if(param.is_open())
  {
    
    cout << "Reading input parameters from file 'coefficients.dat'\n";
    param >> size;
    skipline(param); 
    M.resize(size,size);
    param >> M;
    cout<< M<< endl;
    // puts the matrix in right form //
    for(i=0;i<M.getHeight();i++)
    {
      for(j=0;j<(int) (M.getHeight()-1)/2;j++)
      {
        to=M[i][M.getHeight()-1-j];
        M[i][M.getHeight()-j-1]=M[i][j];
        M[i][j]=to;
      }
    }
    skipline(param);
    param >> inz1;
    skipline(param);
    param >> inz2;  //initial point
    skipline(param);
    param >> inv1;
    skipline(param);
    param >> inv2;  //initial velocity
    skipline(param);
    param >> Q1;
    skipline(param);
    param >> Q2; //angular momenta
    skipline(param);
    param >> del; //step length
    skipline(param);
    param >> Num;
    skipline(param);
    param >> choice_output;
    skipline(param);
    param >> zoom;
    skipline(param);
    param >> speed;
    param.close();
    
  }
  else
  { 
    cout<< "\n --|> Parameter file 'coefficients.dat' not found ! \n"<<endl;
  } 
  double z1 = inz1;
  double z2 = inz2;
  double v1 = inv1;
  double v2 = inv2;
  double th1 = 0;
  double th2 = 0;

   
  MMatrix <double>  cz1(4,1);
  MMatrix <double>  cz2(4,1);
  MMatrix <double>  cv1(4,1);
  MMatrix <double>  cv2(4,1);
  MMatrix <double>  mult(4,1);

  mult[0][0] = .5;
  mult[1][0] = .5;
  mult[2][0] = 1.;
  mult[3][0] = 1.;
  double fz1,fz2,fv1,fv2,tht1,tht2,rat,x1,x2,thv1,thv2,dv1,dv2,dz1,dz2,h11,h12,h22,KE,PE,KEc;
  double th11=0;
  double th12=0;
  double th22=0;
  int size2=(int) (size-1)/2;
  double at = size2;
  double bt = size2;
  eval(at,bt,v1,v2,z1,z2,x1,x2,thv1,thv2,dv1,dv2,dz1,dz2);
  double iKE = pow2(v1)* h11 + 2 * v1 * v2 * h12 + pow2(v2) * h22;
  double iPE = th11 *pow2(Q1) + 2 * th12 * Q1 * Q2 + th22 * pow2(Q2);
  double iE = iPE + iKE;


  int handle; /* handle to name the plotter */
  char *display_type;
  char *bitmap_size;
  if (choice_output==0)
  {
    display_type = "X";/* format of plotter */
  }
  if (choice_output==1)
  {
    display_type = "png";
  }
  if(zoom==1)
  {
    bitmap_size="600x600"; /* size in pixels of plotter */
  }
  if(zoom==2)
  {
    bitmap_size="1024x1024"; /* size in pixels of plotter */
  }
  if(zoom==3)
  {
    bitmap_size="1600x1600"; /* size in pixels of plotter */
  }
  handle=init_plotsession(display_type,bitmap_size);
  pl_space (-1, -1, size, size); /* defines coordinates of the system */
   
  pl_pencolorname ("green");
  pl_fline(0.,0.,0.,size2);
  pl_fline(0.,0.,size2,0.);
  pl_fline(0.,size2,size2,size-1);
  pl_fline(size2,0.,size-1,size2);
  pl_fline(size-1,size2,size-1,size-1);
  pl_fline(size2,size-1,size-1,size-1);
  pl_pencolorname ("brown");
  pl_fcircle(1.25*size/7.,size-0.5*size/7.,0.5);
  pl_fcircle(0.5*size/7.,size-1.25*size/7.,0.5);

  int t;
  for(t=1;t<Num+1;t++)
  {

    fz1 = z1;
    fz2 = z2;
    fv1 = v1;
    fv2 = v2;

    for(a=0;a<4;a++)
    {
      eval(at,bt,v1,v2,z1,z2,x1,x2,thv1,thv2,dv1,dv2,dz1,dz2);
         //   printf("\n  --> %f %f %f %f %f | %f %f %f %f %f %f %f %f %f %f\n",v1,v2,z1,z2,x1,x2,thv1,thv2,dv1,dv2,dz1,dz2,h11,h12,h22);
      cz1[a][0] = del * dz1;
      cz2[a][0] = del * dz2;
      cv1[a][0] = del * dv1;
      cv2[a][0] = del * dv2;
      z1 = fz1 + mult[a][0] * cz1[a][0];
      z2 = fz2 + mult[a][0] * cz2[a][0];
      v1 = fv1 + mult[a][0] * cv1[a][0];
      v2 = fv2 + mult[a][0] * cv2[a][0];
    }

    z1 = fz1 + (cz1[0][0] + 2 * cz1[1][0] + 2 * cz1[2][0] + cz1[3][0]) / (size-1.);
    z2 = fz2 + (cz2[0][0] + 2 * cz2[1][0] + 2 * cz2[2][0] + cz2[3][0]) / (size-1.);
    v1 = fv1 + (cv1[0][0] + 2 * cv1[1][0] + 2 * cv1[2][0] + cv1[3][0]) / (size-1.);
    v2 = fv2 + (cv2[0][0] + 2 * cv2[1][0] + 2 * cv2[2][0] + cv2[3][0]) / (size-1.);
    tht1 = th1 + del * thv1;
    tht2 = th2 + del * thv2;

    eval(at,bt,v1,v2,z1,z2,x1,x2,thv1,thv2,dv1,dv2,dz1,dz2);
    KE = pow2(v1) * h11 + 2 * v1 * v2 * h12 + pow2(v2) * h22;
    PE = th11 * pow2(Q1) + 2 * th12 * Q1 * Q2 + th22 * pow2(Q2);
    KEc = iE - PE;
    rat = 1;
    if(KE > .0001 && t == 100 * (int (t / 100.)))
    {
      rat = KEc / KE;
    }

    v1 = v1 * sqrt(rat);
    v2 = v2 * sqrt(rat);
    pl_pencolorname ("red");
    pl_fpoint(x1,x2);
    if (speed==0 || speed==2)
    {
      if (speed==2)
      {  
           
        start_time = clock();
        while((clock() - start_time) <0.01 * CLOCKS_PER_SEC)
        {
        }
      }
      pl_pencolorname ("white");
      pl_fline(.48*size/7. + .4 * cos(th1 / 2.),size-( 1.25*size/7. + .4 * sin(th1 / 2.)),.52*size/7. + .4 * cos(th1 / 2.), size-(1.25*size/7. + .4 * sin(th1 / 2.)));
      pl_fline(1.25*size/7. + .4 * cos(th2 / 2),size-( .48*size/7. + .4 * sin(th2 / 2)),1.25*size/7. + .4 * cos(th2 / 2), size-(.52*size/7. + .4 * sin(th2 / 2)));
      pl_pencolorname ("blue");
      pl_fline(.48*size/7. + .4 * cos(tht1 / 2.),size- (1.25*size/7. + .4 * sin(tht1 / 2.)),.52*size/7. + .4 * cos(tht1 / 2.), size-(1.25*size/7. + .4 * sin(tht1 / 2.)));
      pl_fline(1.25*size/7. + .4 * cos(tht2 / 2), size- (.48*size/7. + .4 * sin(tht2 / 2)),1.25*size/7. + .4 * cos(tht2 / 2),size- (.52*size/7. + .4 * sin(tht2 / 2)));
    }
      
    th1 = tht1;
    th2 = tht2;
  }

  std::ostringstream s2,s2b,s3;
  s2 << "Initial point:(" << inz1 << "," << inz2 << ")    Angular momenta 1,2: " << Q1 << "," << Q2;
  s2b << "    Initial velocity 1,2: "<< inv1 << ","<< inv2 << "   Step length: "<< del ;

  std::string name;
  name=s2.str();
  pl_fontname ("Helvetica");
  pl_fontsize(-1);
  pl_pencolorname ("black");
  pl_fmove(-0.9,-0.6);
  pl_label(name.c_str());
  pl_fmove(-0.9,-0.9);
  name=s2b.str();
  pl_label(name.c_str());
  pl_fmove(4.5,6.5);
  s3  << " No Iterations :"<< Num;
  name=s3.str();
  pl_label(name.c_str());
  end_plotsession(handle);

}



void eval(double at,double bt,double v1,double v2,double z1,double z2,double &x1,double &x2,double &thv1,double &thv2,double &dv1,double &dv2,double &dz1,double &dz2)
{

  int a,b;
  double f = 0;
  double f1 = 0;
  double f2 = 0;
  double f11 = 0;
  double f12 = 0;
  double f22 = 0;
  double f111 = 0;
  double f112 = 0;
  double f122 = 0;
  double f222 = 0;
  double ext,ad,bd;
  double h11,h12,h22;
  double Mabext,exta,extb,extastep,extbstep,extb0;
  exta=exp((-at-1) * z1);
  extb0=exp((-bt-1) * z2);
  extastep=exp(z1);
  extbstep=exp(z2);
  for (a = 0;a<size;a++)
  {
    exta=exta*extastep;
    extb=extb0;
    ad = a - at;
    for (b = 0;b<size;b++)
    {
   
      bd = b - bt;
      extb=extb*extbstep;
      ext = exta*extb; //exp(ad * z1 + bd * z2);
      Mabext = M[a][b] * ext;
      f = f + Mabext;
      f1 = f1 + ad * Mabext;
      f2 = f2 + bd * Mabext;
      f11 = f11 + pow2(ad) * Mabext;
      f12 = f12 + ad * bd * Mabext;
      f22 = f22 + pow2(bd) * Mabext;
      f111 = f111 + pow3(ad)* Mabext;
      f112 = f112 + pow2(ad) * bd * Mabext;
      f122 = f122 + ad * pow2(bd) * Mabext;
      f222 = f222 + pow3(bd) * Mabext;
    }
  }
//  x1 = at + f1 / f;
//  x2 = bt + f2 / f;
  x1=at+f2/f;  //conjugate
  x2=bt-f1/f;
  h11 = f11 / f - pow2(f1 / f);
  h12 = f12 / f - f1 * f2 / pow2(f);
  h22 = f22 / f - pow2(f2 / f);
  double H111 = f111 / f + 2 * pow3(f1 / f) - 3 * f11 * f1 / pow2(f);
  double H112 = f112 / f + 2 * pow2(f1)* f2 / pow3(f) - f11 * f2 / pow2(f) - 2 * f12 * f1 / pow2(f);
  double H122 = f122 / f + 2 * pow2(f2) * f1 / pow3(f) - f22 * f1 / pow2(f) - 2 * f12 * f2 / pow2(f);
  double H222 = f222 / f + 2 * pow3(f2 / f) - 3 * f22 * f2 / pow2(f);
  double D = h11 * h22 - pow2(h12);
  double Ht11 = h22 / D;
  double Ht22 = h11 / D;
  double Ht12 = -h12 / D;
  thv1 = Ht11 * Q1 + Ht12 * Q2;
  thv2 = Ht12 * Q1 + Ht22 * Q2;
  dz1 = v1;
  dz2 = v2;
  double P1 = -.5 * (H111 * (pow2(v1) - pow2(thv1)) + 2 * H112 * (v1 * v2 - thv1 * thv2) + H122 * (pow2(v2) - pow2(thv2)));
  double P2 = -.5 * (H112 * (pow2(v1) - pow2(thv2)) + 2 * H122 * (v1 * v2 - thv1 * thv2) + H222 *(pow2(v2)- pow2(thv2)));
  dv1 = Ht11 * P1 + Ht12 * P2;
  dv2 = Ht12 * P1 + Ht22 * P2;
}






int init_plotsession(char *display_type, char *bitmap_size)
{
  int handle, return_value;
  return_value=pl_parampl("BITMAPSIZE", bitmap_size);
  if(return_value<0)
  {
    fprintf(stderr,"The plotter does not accept the parameters specified.\n");
    exit(1);
  }
  handle=pl_newpl(display_type,stdin, stdout, stderr);

  if(handle<0)
  {
    fprintf(stderr,"Can't create plotter.\n");
    exit(1);
  }
  return_value=pl_selectpl(handle);
  if(return_value<0)
  {
    fprintf(stderr,"Plotter does not exist\n");
    exit(1);
  }
  return_value=pl_openpl();
  if(return_value<0)
  {
    fprintf(stderr,"Plotter can't be opened.\n");
    exit(1);
  }
  return handle;
}


/******************************************************************************/

int end_plotsession(int handle)

    /* closes the plotter -> 0 if OK */
{
  int err_status=0, return_value;
  return_value= pl_closepl();
  if(return_value<0)
  {
    fprintf(stderr,"Problem to close the plotter\n");
    /* no exit, because we try to delete the plotter */
    err_status +=1;
  }
  return_value=pl_selectpl(0);
  if(return_value<0)
  {
    fprintf(stderr,"Default Plotter could not be selected!\n");
    err_status +=1<<1;
  }
  /* deleting our plotter */
  return_value=pl_deletepl (handle);
  if(return_value<0)
  {
    fprintf(stderr,"Cannot delete plotter !\n");
    err_status +=1<<2;
  }
  return err_status;
}


//utility procedure to handle input file//
void skipline(ifstream & mystream)
{
  mystream.ignore(100,'\n');
}
